#!/usr/bin/env python

###################################
# filename: knpsdp.py
# author: Leo Liberti
# purpose: SDP relaxations of KNP formulation
#          (Maculan, Michelon & Smith) and KNP feasibility
# source: Python 2.7
# requirements: Python >= 2.7.11,
#               pip install numpy scipy matplotlib algopy cvxopt picos
#               mosek, cplex (python $i install)
###################################

############# SECT:IMPORTS #############
# basic
import sys         # system calls (.exit, .argv)
import re          # regular expression library, used in reading input files
import os.path
import time
import math        # standard math functions 
import numpy as np # the numeric library, with linear algebra calls
#from numpy import *
import scipy
from scipy import optimize # to use the optimization solvers in scipy 
from collections import OrderedDict # used to nice-print adjacency lists
import itertools   # find initial clique in DVOP order greedy alg
import operator    # operator.itemgetter() to find max in tuple components
import types
import warnings

# plots
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import proj3d
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches

# CPLEX
import cplex
from cplex.exceptions import CplexError

# SDP
import cvxopt as cvx # the CvxOpt library, with various solvers including SDP
import picos as pic  # the modelling layer for CvxOpt

# options
np.set_printoptions(precision=2, suppress=True) # print precision for numpy

############# SECT:GLOBALS ###############

# set number of dimensions
global NumberOfDimensions
NumberOfDimensions = 3

# this is a technical hack to plot in 3D and update the plot after mouse ops
#   no need to change / update this; see
#   stackoverflow.com/questions/10374930/matplotlib-annotating-a-3d-scatter-plot
global labels_and_points
labels_and_points = []

# this is a tolerance for zero, used in next* computations
global myZero
myZero = 1e-6

# this is an upper bound on weights
global myInf
myInf = 1e+25

# iterations of DDP iterative method
global DDPItn
DDPItn = 5

# maximum iteration for local NLP solvers (<0 to accept solver's default)
nlpMaxIter = -1

# SDP solver tolerance (<0 to accept solver's default)
solverTolerance = 1.0e-3

    
############# SECT:BASIC_METHODS ############

### generate an Achlioptas m x n random projector matrix
def achlioptasProjector(m, n):
    A = np.zeros((m,n))
    for i in range(m):
        for j in range(n):
            aij = achlioptas_sample()
            if aij != 0:
                A[i,j] = aij
    return A

### generate a samples from an Achlioptas distribution
def achlioptas_sample():
    ret = 0
    uniformsample = np.random.uniform(0,1)
    if uniformsample < 1.0/6.0:
        ret = -1
    elif uniformsample > 5.0/6.0:
        ret = 1
    else:
        ret = 0
    return ret

## compute the Gram matrix from the distance matrix
def dist2Gram(EDM):
    n = EDM.shape[0]
    J = np.identity(n) - (1.0/n)*np.ones((n,n))
    B = -0.5 * np.dot(J,np.dot(np.square(EDM), J))
    return B

## return the distance matrix of a realization
##   p is the norm order
def distanceMatrix(x, p=2):
    n = len(x[:,0])
    EDM = np.zeros((n,n))
    for u in range(n-1):
        for v in range(u+1,n):
            EDM[u,v] = np.linalg.norm(np.subtract(x[u,:],x[v,:]), ord=p)
            EDM[v,u] = EDM[u,v]
    return EDM

## matrix without the j-th column
def exceptColumn(A, j):
    return np.delete(A,j,1)

## matrix without the i-th row
def exceptRow(A, i):
    return np.delete(A,i,0)

### generate a Gaussian m x n random projector matrix 
def gaussianProjector(m, n):
    return (1.0/math.sqrt(m)) * np.random.normal([0],[[1]],(m,n))

## is given matrix PSD?
def isPSD(A):
  eigvals, eigvects = np.linalg.eigh(A)
  return np.all(eigvals > -myZero)

## pad an np.array with zeros up to length K
def lift(a, K):
    t = K - len(a)
    if t > 0:
        b = np.pad(a, (0,t), "constant")
    else:
        b = a
    return b

## pad a list with zeros up to length K
def liftList(a, K):
    t = K - len(a)
    if t > 0:
        a.extend([0]*t)
    return a

## count near-zeros of a matrix
def numNearZeros(A):
    ret = 0
    (m,n) = A.shape
    for i in range(m):
        for j in range(n):
            if abs(A[i,j]) <= myZero:
                ret = ret + 1
    return ret

## (random) r-diagonal matrices, see [Barvinok DCG 1995, example 4.1]
def rDiagonal(n,r):
    F = np.identity(n)
    for i in range(n):
        for j in range(r):
            if i+j+1 < n:
                rndeps = np.random.uniform(-1.0/r,1.0/r)
                F[i,i+j+1] = rndeps
                F[i+j+1,i] = rndeps
    return F

## range without the i-th element
def rangeBut(r, i):
    R = range(r)
    return R[:i] + R[i+1:]

## relative distance (closeness) between two vectors
def relativeCloseness(x,y):
    nx = np.linalg.norm(x)
    ny = np.linalg.norm(y)
    sx = (1/nx) * x
    sy = (1/ny) * y
    angle = abs(np.dot(sx,sy))
    anglefraction = angle / math.pi
    normfraction = min(nx,ny) / max(nx,ny)
    return (anglefraction, normfraction)

## random square positive definite matrix
def rndPositiveDefinite(n):
    F = np.identity(n)
    for i in range(n):
        F[i,i] = np.random.uniform(1.0,5.0)
    P = np.zeros((n,n))
    for i in range(n):
        for j in range(i+1,n):
            P[i,j] = np.random.uniform(1.0,5.0)
    return np.dot(P.T,np.dot(F,P))

## return the squared distance matrix of a realization
def squareDistanceMatrix(x):
    return np.square(distanceMatrix(x))

############# SECT:COMBINATORIAL #########

## the IsoMap method with a full distance matrix
def IsoMapEDM(A):
    B = dist2Gram(A)
    x = PCA(B, NumberOfDimensions)
    return x

## perform PCA
def PCA(B, K = "None"):
    n = B.shape[0]
    x = factor(B) # returns cols sorted by decreasing eigval
    if type(K) == types.StringType:
        K = n
    if K < n:
        # only first K columns (corresp to largest eigvals)
        x = x[:,0:K]
    return x

## factor a square matrix 
def factor(A):
    n = A.shape[0]
    (evals,evecs) = np.linalg.eigh(A)
    evals[evals < 0] = 0  # closest SDP matrix
    X = evecs #np.transpose(evecs)
    sqrootdiag = np.eye(n)
    for i in range(n):
        sqrootdiag[i,i] = math.sqrt(evals[i])
    X = X.dot(sqrootdiag)
    return np.fliplr(X) # return sorted by increasing eigval
    return X

############# SECT:FORMULATION_BASED #############

## build and solve SDP relaxation of KNP in PICOS
##   formulation by [Maculan, Michelon, Smith]
def MaculanMichelonSmithSDP(n, K):
    t0 = time.time()
    KNP = pic.Problem()
    IdK = cvx.matrix(np.identity(K))

    # decision variables
    X = KNP.add_variable('X', (n,n), vtype='symmetric', lower=-1, upper=1) 
    x = KNP.add_variable('x', (n,K), lower=-1, upper=1)
    A = KNP.add_variable('A', (n,n), vtype='symmetric', lower=0, upper=1)
    a = KNP.add_variable('a', (n,1), vtype='continuous', lower=0, upper=1)
    #a = KNP.add_variable('a', (n,1), vtype='integer', lower=0, upper=1)

    # objective function
    KNP.set_objective('max',pic.sum([a[i] for i in range(n)], 'i', 'range(n)'))
    #KNP.set_objective('min',( cvx.matrix(rDiagonal(n,K)) | X ))
    #KNP.set_objective('min',( cvx.matrix(rDiagonal(n,K)) | X ) - pic.sum([a[i] for i in range(n)], 'i', 'range(n)'))

    # constraints
    KNP.add_constraint(sum(X[i,i] for i in range(n)) <= n)
    for i in range(n):
        KNP.add_constraint(X[i,i] == a[i])
        KNP.add_constraint(A[i,i] == a[i])
        for j in range(n):
            if i < j:
                KNP.add_constraint(X[i,i] + X[j,j] - 2*X[i,j] >= A[i,j])
                KNP.add_constraint(A[i,j] >= a[i] + a[j] - 1)
                KNP.add_constraint(A[i,j] <= a[i])
                KNP.add_constraint(A[i,j] <= a[j])
                
    KNP.add_constraint(( ( IdK & x.T ) // ( x & X ) ) >> 0)
    KNP.add_constraint(( ( 1 & a.T ) // ( a & A ) ) >> 0)
    t1 = time.time()
    # solve and get solution
    KNP.solve()
    t2 = time.time()
    print "MaculanMichelonSmithSDP: modelling time =", t1-t0, "; solving time =",t2-t1
    Xv = np.array(X.value)
    xv = np.array(x.value)
    Av = np.array(A.value)
    av = np.array(a.value)
    objv = KNP.obj_value()
    return (objv, Xv, xv, Av, av)

## build and solve SDP relaxation of KNP feasibility in PICOS
##   feasibility-only: can n balls fit around a central ball in R^K?
def KNPFeasibilitySDP(n, K = NumberOfDimensions):
    t0 = time.time()
    KNP = pic.Problem()
    IdK = cvx.matrix(np.identity(K))

    # decision variables
    X = KNP.add_variable('X', (n,n), vtype='symmetric', lower=-1, upper=1) 
    x = KNP.add_variable('x', (n,K), lower=-1, upper=1)
    a = KNP.add_variable('a', (1,1), vtype='continuous', lower=0)

    # objective function
    KNP.set_objective('max', a)

    # constraints
    for i in range(n):
        KNP.add_constraint(X[i,i] == 1)
        for j in range(n):
            if i < j:
                KNP.add_constraint(X[i,i] + X[j,j] - 2*X[i,j] >= a)
                
    KNP.add_constraint(( ( IdK & x.T ) // ( x & X ) ) >> 0)
    #KNP.add_constraint(X >> 0)
    t1 = time.time()
    # solve and get solution
    KNP.solve()
    t2 = time.time()
    print "KNPFeasibilitySDP: modelling time =", t1-t0, "; solving time =",t2-t1
    Xv = np.array(X.value)
    objv = KNP.obj_value()
    av = np.array(a.value)
    return (objv, Xv, av[0,0])


####### MAIN #######


if len(sys.argv) < 4:
    print sys.argv[0], ": need cmd line, syntax is"
    print "  ", sys.argv[0], "<feas|mms95> n K"
    exit()

formulation = sys.argv[1]
n = int(sys.argv[2])
K = int(sys.argv[3])

print "solving", formulation, "(SDP relax.): n =", n, "and K =", K

if formulation == "feas":
    (obj,X,a) = KNPFeasibilitySDP(n, K)
elif formulation == "mms95":
    (obj,X,x,A,a) = MaculanMichelonSmithSDP(n, K)
else:
    exit('available formulations are feas and mms95')

# for i in range(n):
#     for j in range(n):
#         if i <= j:
#             print "{0:.2f}".format(X[i,i] + X[j,j] - 2*X[i,j]),
#         else:
#             print "    ",
#     print

if formulation == "feas":
    x = PCA(X, K)
    fn = "knp-" + str(n) + "_" + str(K) + "-xbar.dat"
    f = open(fn, "w")
    print >> f, "param n :=", n, ";"
    print >> f, "param K :=", K, ";"
    print >> f, "param abar :=", a, ";"
    print >> f, "param xbar :="
    for i in range(n):
        for k in range(K):
            print >> f, " ", i+1, k+1, " ", x[i,k]
    print >> f, ";"
    f.close()
    print "optimal objective:", obj, ", optimal alpha:", a;

elif formulation == "mms95":
    print "optimal objective:", obj, ", optimal sum(slacks):", (sum(a))[0]
    #print np.linalg.matrix_rank(X), np.linalg.matrix_rank(A)
